#!/usr/bin/env python

from cortix.src.module import Module
from cortix.examples.dummy_module import DummyModule

def test_module_init():
    # Initialize the module
    m = DummyModule()

    # get ports
    p1 = m.get_port('test-1')
    p2 = m.get_port('test-2')

    # Output the list of ports
    print(m.ports)
    assert(len(m.ports) == 2)

if __name__ == "__main__":
    test_module_init()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import sys
import logging
from threading import Thread
import pickle

from mpl_toolkits.mplot3d import Axes3D
import matplotlib
matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt

from cortix.src.module import Module

class DataPlot(Module):

    def __init__(self):

        super().__init__()

        self.same_axes = False
        self.dpi = 200

        self.xlabel = 'x'
        self.ylabel = 'y'
        self.zlabel = 'z'
        self.title = None

        self.log = logging.getLogger('cortix')
        self.debug = False
        self.print_freq = 10

        self.data = dict()

        self.state = self.data

    def run(self, state_comm=None, idx_comm=None):
        '''
        Spawn a thread to handle each port connection.
        '''

        threads = []

        for port in self.ports:
            thread = Thread(target=self.recv_data, args=(port,))
            thread.start()
            threads.append(thread)

        for t in threads:
            t.join()

        self.plot_data()

        if state_comm:
            try:
                pickle.dumps(self.state)
            except pickle.PicklingError:
                state_comm.put((idx_comm,None))
            else:
                state_comm.put((idx_comm,self.state))

        return

    def recv_data(self, port):
        '''
        Keep listening on the port and receiving data.
        '''

        data = []

        i = 1
        while True:
            d = self.recv(port)
            if self.debug and i % self.print_freq == 0:
                self.log.info('DataPlot::'+port.name+' received: {}'.format(d))
            if isinstance(d, str) and d == "DONE":
                self.data[port.name] = data
                sys.exit(0) # kill thread
            i += 1

            data.append(d)

    def plot_data(self):

        if self.same_axes:
            fig = plt.figure(1)
            ax = None

        for (key,data) in self.data.items():

            x = [i[0] for i in data]
            y = [i[1] for i in data]

        # 2D-Plot
            if data and len(data[0]) == 2:
                if not self.same_axes:
                    fig = plt.figure(key)
                plt.xlabel(self.xlabel)
                plt.ylabel(self.ylabel)
                plt.title(self.title)
                plt.plot(x, y)

        # 3D-Plot
            elif data and len(data[0]) == 3:

                if self.same_axes and ax is None:
                    ax = fig.add_subplot(111, projection='3d')
                    ax.set_xlabel(self.xlabel)
                    ax.set_ylabel(self.ylabel)
                    ax.set_zlabel(self.zlabel)
                    ax.set_title(self.title)

                if not self.same_axes:
                    fig = plt.figure(key)
                    ax = fig.add_subplot(111, projection='3d')
                    ax.set_xlabel(self.xlabel)
                    ax.set_ylabel(self.ylabel)
                    ax.set_zlabel(self.zlabel)
                    ax.set_title(self.title)

                ax.plot(x, y, [i[2] for i in data])

                if not self.same_axes:
                    plt.savefig('{}.png'.format(key), dpi=self.dpi)

        if self.same_axes:
            plt.savefig('{}.png'.format(key.split(':')[0]), dpi=self.dpi)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from cortix.src.module import Module
from cortix.src.cortix_main import Cortix

from cortix.examples.dataplot import DataPlot
from cortix.examples.droplet import Droplet
from cortix.examples.vortex import Vortex

'''
This example uses three modules instantiated many times in two different networks.
Each network configuration uses a different amount of module instances and a different
network topology. This example can be executed with MPI (if mpi4py is available) or
with the Python multiprocessing library. These choices are made by variables listed
below in the executable portion of this run file.

Single Plot
-----------

The first network case is named "single plot". Here one DataPlot module is connected
to all Droplet modules. To run this case using MPI you should compute the number of
processes as follows:

    `nprocs = n_droplets + 1 vortex + 1 data_plot + 1 cortix`

then issue the MPI run command as follows (replace `nprocs` with a number):

     `mpiexec -n nprocs run_droplet.py`

To run this case with the Python multiprocessing library, just run this file at the
command line as

     `run_droplet.py`

Multiple Plot
-------------

The second network case is named "multiple plot". Here each Droplet is connected to an
instance of the DataPlot module, therefore many more nodes are added to the network
when compared to the first network case. To run this case using MPI compute

    `nprocs = 2*n_droplets + 1 vortex + 1 cortix`

then issue the MPI run command as follows (replace `nprocs`:

    `mpiexec -n nprocs run_droplet.py`

To run this case with the Python multiprocessing library, just run this file at the
command line as

    `run_droplet.py`
'''

if __name__ == '__main__':

    # Configuration Parameters
    use_single_plot = True  # True for a single plot output
                            # False for multiple plot files and network
    use_mpi         = False # True for MPI; False for Python multiprocessing

    plot_vortex_profile = False # True may crash the X server.

    n_droplets = 5
    end_time   = 300
    time_step  = 0.1

    cortix = Cortix(use_mpi=use_mpi)

    # Network for a single plot case
    if use_single_plot:

        # Vortex module (single).
        vortex = Vortex()
        cortix.add_module(vortex)
        vortex.show_time = (True,100)
        vortex.end_time = end_time
        vortex.time_step = time_step
        if plot_vortex_profile:
            vortex.plot_velocity()

        # DataPlot module (single).
        data_plot = DataPlot()
        cortix.add_module(data_plot)
        data_plot.title = 'Droplet Trajectories'
        data_plot.same_axes = True
        data_plot.dpi = 300

        for i in range(n_droplets):

            # Droplet modules (multiple).
            droplet = Droplet()
            cortix.add_module(droplet)
            droplet.end_time = end_time
            droplet.time_step = time_step
            droplet.bounce = False
            droplet.slip = False

            # Network port connectivity (connect modules through their ports)
            droplet.connect('external-flow', vortex.get_port('fluid-flow:{}'.format(i)))
            droplet.connect('visualization', data_plot.get_port('viz-data:{:05}'.format(i)))

    # Network for a multiple plot case
    if not use_single_plot:

        # Vortex module (single).
        vortex = Vortex()
        cortix.add_module(vortex)
        vortex.show_time = (True,100)
        vortex.end_time = end_time
        vortex.time_step = time_step
        if plot_vortex_profile:
            vortex.plot_velocity()

        for i in range(n_droplets):

            # Droplet modules (multiple).
            droplet = Droplet()
            cortix.add_module(droplet)
            droplet.end_time = end_time
            droplet.time_step = time_step
            droplet.bounce = False
            droplet.slip = False

            # DataPlot modules (multiple).
            data_plot = DataPlot()
            cortix.add_module(data_plot)
            data_plot.title = 'Droplet Trajectory '+str(i)
            data_plot.dpi = 300

            # Network port connectivity (connect modules through their ports)
            droplet.connect('external-flow', vortex.get_port('fluid-flow:{}'.format(i)))
            droplet.connect('visualization', data_plot.get_port('viz-data:{:05}'.format(i)))

    cortix.draw_network('network.png')

    cortix.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

import os
import logging
import time
import datetime
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from multiprocessing import Process, Queue
from cortix.src.module import Module

class Cortix:
    '''
    The main Cortix class definition:
    1. Create the object
    2. Add modules
    3. Run the simulation
    '''

    def __init__(self, use_mpi=False, splash=True):

        self.use_mpi = use_mpi

        self.comm = None
        self.rank = None
        self.size = None

        self.splash = splash

        # Fall back to multiprocessing if mpi4py is not available
        if self.use_mpi:
            try:
                from mpi4py import MPI
                self.comm = MPI.COMM_WORLD
                self.rank = self.comm.Get_rank()
                self.size = self.comm.size
            except ImportError:
                self.use_mpi = False

        # Setup the global logger 
        self.__create_logger()

        # Setup the network graph
        if self.rank == 0 or not self.use_mpi:
            self.nx_graph = None

        # Modules storage
        self.modules = list()

        # Done
        if self.rank == 0 or not self.use_mpi:

            if self.splash:
                self.log.info('Created Cortix object %s', self.__get_splash(begin=True))
            else:
                self.log.info('Created Cortix object')

            self.wall_clock_time_start = time.time()

            self.wall_clock_time_end = self.wall_clock_time_start
            self.end_run_date = datetime.datetime.today().strftime('%d%b%y %H:%M:%S')

        return

    def __del__(self):
        '''
        Note: by the time the body of this function is executed, the machinery of
        variables may have been deleted already. For example, `logging` is no longer
        there; do the least amount of work here.
        '''

        if self.rank == 0 or not self.use_mpi:

            if self.splash:
                print('Destroyed Cortix object on '+self.end_run_date+
                        self.__get_splash(begin=False))
            else:
                print('Destroyed Cortix object on '+self.end_run_date)

            print('Elapsed wall clock time [s]: '+
                    str(round(self.wall_clock_time_end-self.wall_clock_time_start,2)))

    def add_module(self, m):
        '''
        Add a module to the Cortix object

        `m`: An instance of a class that inherits from the Module base class
        '''

        assert isinstance(m, Module), 'm must be a module'
        if m not in self.modules:
            m.use_mpi = self.use_mpi
            self.modules.append(m)

    def run(self):
        '''
        Run the Cortix simulation with MPI if available o.w. fallback to multiprocessing
        '''

        # Running under MPI
        #------------------
        if self.use_mpi:

            # Synchronize in the beginning
            assert self.size == len(self.modules) + 1,\
                'Incorrect number of processes (Required %r, got %r)'%\
                (len(self.modules) + 1, self.size)
            self.comm.Barrier()

            # Assign an mpi rank to all ports of a module using the module list index
            for m in self.modules:
                rank = self.modules.index(m)+1
                for port in m.ports:
                    port.rank = rank

            # Assign a unique port id to all ports
            i = 0
            for mod in self.modules:
                for port in mod.ports:
                    port.use_mpi = self.use_mpi
                    port.id = i
                    i += 1

            # Parallel run module in MPI
            if self.rank != 0:
                mod = self.modules[self.rank-1]
                self.log.info('Launching Module {}'.format(mod))
                mod.run()

            # Synchronize at the end
            self.comm.Barrier()

        # Running under Python multiprocessing
        #-------------------------------------
        else:

            # Parallel run all modules in Python multiprocessing
            processes = list()
            modules_new_state = Queue()

            for mod in self.modules:
                self.log.info('Launching Module {}'.format(mod))
                p = Process(target=mod.run,
                        args=( modules_new_state, self.modules.index(mod)) )
                processes.append(p)
                p.start()

            # Update module states in the parent process
            for q in range(len(processes)):
                (mod_idx, new_state) = modules_new_state.get()
                self.modules[mod_idx].state = new_state
                self.log.info('Module {} getting new state'.format(self.modules[mod_idx]))

            # Synchronize at the end
            for p in processes:
                p.join()

        # Record time at the end of the run method
        if self.rank == 0 or not self.use_mpi:
            self.end_run_date = datetime.datetime.today().strftime('%d%b%y %H:%M:%S')
            self.wall_clock_time_end = time.time()

        return

    def get_network(self):
        '''
        Constructs and returns a networkx graph representation of the module network.
        '''
        if not self.use_mpi or self.rank == 0:
            if not self.nx_graph:
                g = nx.MultiGraph()
                for mod_one in self.modules:
                    class_name = mod_one.__class__.__name__
                    index = self.modules.index(mod_one)
                    mod_one_name = "{}_{}".format(class_name, index)
                    for mod_two in self.modules:
                        if mod_one != mod_two:
                            for port in mod_one.ports:
                                for p2 in mod_two.ports:
                                    if id(port.connected_port) == id(p2):
                                        mod_two_name = "{}_{}".format(mod_two.__class__.__name__, self.modules.index(mod_two))
                                        if mod_two_name not in g or mod_one_name not in g.neighbors(mod_two_name):
                                            g.add_edge(mod_one_name, mod_two_name)
                self.nx_graph = g
            return self.nx_graph

    def draw_network(self, file_name='network.png', dpi=220):
        '''
        Draws the networkx Module network graph using matplotlib

        `file_name`: The resulting network diagram output file name
        `dpi`: dpi used for generating the network image
        '''
        if self.use_mpi and self.rank != 0:
            return

        g = self.nx_graph if self.nx_graph else self.get_network()
        colors = ["blue", "red", "green", "teal"]
        class_map = {}
        color_map = {}
        for node in g.nodes():
            class_name = "_".join(node.split("_")[:-1])
            if class_name not in class_map:
                class_map[class_name] = colors[len(class_map) % len(colors)]
            color_map[node] = class_map[class_name]
        f = plt.figure()
        pos = nx.spring_layout(g, k=0.15, iterations=20)
        nx.draw(g, pos, node_color=[color_map[n] for n in g.nodes], ax=f.add_subplot(111), linewidths=0.01)
        patches = []
        for c in class_map:
            patch = mpatches.Patch(color=class_map[c], label=c)
            patches.append(patch)
        plt.legend(handles=patches)
        f.savefig(file_name, dpi=dpi)

    def __create_logger(self):
        '''
        A helper function to setup the logging facility used in the constructor.
        '''

        # File removal
        if self.rank == 0 or not self.use_mpi:
            if os.path.isfile('cortix.log'):
                os.system('rm -rf cortix.log')

        # Sync here to allow for file removal
        if self.use_mpi:
            self.comm.Barrier()

        self.log = logging.getLogger('cortix')

        self.log.setLevel(logging.DEBUG)

        file_handler = logging.FileHandler('cortix.log')
        file_handler.setLevel(logging.DEBUG)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)

        # Formatter added to handlers
        if self.use_mpi:
            fs = '[rank:{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s'.format(self.rank)
        else:

            fs = "[{}] %(asctime)s - %(name)s - %(levelname)s - %(message)s".format(os.getpid())

        formatter = logging.Formatter(fs)
        file_handler.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        # Add handlers to logger
        self.log.addHandler(file_handler)
        self.log.addHandler(console_handler)

    def __get_splash(self, begin=True):

        splash = \
        '_____________________________________________________________________________\n'+\
        '      ...                                        s       .     (TAAG Fraktur)\n'+\
        '   xH88"`~ .x8X                                 :8      @88>\n'+\
        ' :8888   .f"8888Hf        u.      .u    .      .88      %8P      uL   ..\n'+\
        ':8888>  X8L  ^""`   ...ue888b   .d88B :@8c    :888ooo    .     .@88b  @88R\n'+\
        'X8888  X888h        888R Y888r ="8888f8888r -*8888888  .@88u  ""Y888k/"*P\n'+\
        '88888  !88888.      888R I888>   4888>"88"    8888    ''888E`    Y888L\n'+\
        '88888   %88888      888R I888>   4888> "      8888      888E      8888\n'+\
        '88888 `> `8888>     888R I888>   4888>        8888      888E      `888N\n'+\
        '`8888L %  ?888   ! u8888cJ888   .d888L .+    .8888Lu=   888E   .u./"888&\n'+\
        ' `8888  `-*""   /   "*888*P"    ^"8888*"     ^%888*     888&  d888" Y888*"\n'+\
        '   "888.      :"      "Y"          "Y"         "Y"      R888" ` "Y   Y"\n'+\
        '     `""***~"`                                           ""\n'+\
        '                             https://cortix.org                              \n'+\
        '_____________________________________________________________________________'

        if begin:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                             L A U N C H I N G                               \n'

        else:
            message = \
            '\n_____________________________________________________________________________\n'+\
            '                           T E R M I N A T I N G                             \n'

        return message + splash

if __name__ == '__main__':
    c = Cortix()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

This Specie class is to be used with other classes in plant-level process modules.

NB: Species is always used either in singular or plural cases, the class
    named here reflects one species. If many species are used in an external
    context, the species object name can be used without conflict.

For unit testing do at the linux command prompt:
    python specie.py

NB: The Specie() class encapsulates either the molecular or empirical chemical
    formula of a compound.
    This is done as follows. Say MAO2 is either a molecular or empirical chemical
    formula of a ficticious compound denoting minor actinides dioxide. The list
    of atoms is given as follows:

    ['0.49*Np-237', '0.42*Am-241', '0.08*Am-243', '0.01*Cm-244', '2.0*O-16']

    note the MA forming nuclides add to 1 = 0.49 + 0.42 + 0.08 + 0.01. Therefore
    the number of atoms in this compound is 3. 1 MA "atom" and 2 O.
    Note that the total number of "atoms" is obtained by summing all multipliers:
    0.49 + 0.42 + 0.08 + 0.01 + 2.0.
    The nuclide is indicated by the element symbol followed by a dash and the
    atomic mass number. Here the number of nuclide types is 5 (self._nNuclideTypes).

    The numbers preceeding the nuclide symbol before the * will be referred to as
    multipliers. The sum of the multipliers will add to the number of "atoms" in
    the formula. WARNING: a multiplier could be in the format 0.00e-00. In this
    case a hiphen may appear twice, e.g.: 1.549e-09*U-233

    Other forms can be used for common true species

    ['Np-237', '2.0*O-16'] or ['Np-237', 'O-16', 'O-16'] or [ '2*H', 'O' ] or
    [ 'H', 'O', 'H' ]  etc...

    This code will calculate the molar mass of any species with a given valid
    atom list using a provided periodic table of chemical elements. The user
    can also reset the value of the molar mass with a setter method.

Sat May  9 21:40:48 EDT 2015 created; vfda
'''
#*********************************************************************************
import os
import sys

from cortix.support.periodictable import ELEMENTS
#*********************************************************************************

class Specie:
    '''
    todo: phase should not be here; concentrations should not be here
          only molar quantities should be here
          see the Phase container
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 name='null',
                 formula_name='null',
                 phase='null',
                 atoms=list(),
                 molarCC=0.0,      # default unit: M (mole/L)
                 massCC=0.0,      # default unit: g/L
                 flag=None):

        assert isinstance(name, str), 'oops not string.'
        self._name = name

        assert isinstance(formula_name, str), 'oops not string.'
        self.__formula_name = formula_name

        assert isinstance(phase, str), 'oops not string.'
        self._phase = phase

        assert isinstance(atoms, list), 'oops not list.'
        self._atoms = atoms

        self._flag = flag  # flag can be any type

        self._molarMass = 0.0
        self._molarHeatPwr = 0.0
        self._molarGammaPwr = 0.0
        self._molarRadioactivity = 0.0

        self._molarMassUnit = 'g/mole'

        self._molarHeatPwrUnit = 'W/mole'
        self._molarGammaPwrUnit = 'W/mole'
        self._molarRadioactivityUnit = 'Ci/mole'

        self._molarRadioactivityFractions = list()

        self._molarCCUnit = 'mole/L'
        self._massCCUnit = 'g/L'

        self.__UpdateMolarMass()

        if self._molarMass == 0.0:
            self._molarCC = 0.0
            self._massCC = 0.0
            return

        assert isinstance(molarCC, float), 'oops not a float.'
        assert molarCC >= 0.0, 'oops negative value.'
        self._molarCC = molarCC
        self._massCC = molarCC * self._molarMass

        assert isinstance(massCC, float), 'oops not a float.'
        assert massCC >= 0.0, 'oops negative value.'
        if self._massCC == 0.0:
            self._massCC = massCC
            self._molarCC = massCC / self._molarMass

        return

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def GetName(self):

        '''
        Returns the empirical name of the species. For example, "water".

        Parameters
        ----------
        empty

        Returns
        -------
        name: str

        '''

        return self._name

    def SetName(self, n):

        '''
        Sets the empirical  name of the species to n.

        Parameters
        ----------
        n: str

        Returns
        -------
        empty
        '''

        self._name = n
    name = property(GetName, SetName, None, None)

    def GetFormulaName(self):

        '''
        Returns the formulaic name of the compound. For example, "Dihydrogen
        monoxide".

        Parameters
        ----------
        empty

        Returns
        -------
        self.__formula_name: str

        '''

        return self.__formula_name

    def SetFormulaName(self, f):

        '''
        Sets the formulaic name to f.

        Parameters
        ----------
        empty

        Returns
        -------
        empty
        self.__formula_name: str
        '''

        self.__formula_name = f
    formula_name = property(GetFormulaName, SetFormulaName, None, None)

    def GetPhase(self):

        '''
        Returns the phase history of the species.

        Parameters
        ----------
        empty

        Returns
        -------
        phase: dataFrame
        '''

        return self._phase

    def SetPhase(self, p):

        '''
        Sets the phase history to p.

        Parameters
        ----------
        p: dataFrame

        Returns
        -------
        empty
        '''

        self._mass = p
    phase = property(GetPhase, SetPhase, None, None)

    def GetMolarMass(self):

        '''
        Returns the numerical value for the molar mass of the species. Units
        are given by molarMassUnit.

        Parameters
        ----------
        empty

        Returns
        -------
        molarMass: float
        '''

        return self._molarMass

    def SetMolarMass(self, v):

        '''
        Sets the molar mass of the species equal to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._molarMass = v
    molarMass = property(GetMolarMass, SetMolarMass, None, None)

    def GetMolarMassUnit(self):

        '''
        Returns the unit used to measure the molar mass of the species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarMassUnit: str
        '''

        return self._molarMassUnit

    def SetMolarMassUnit(self, v):

        '''
        Sets the unit used to measure the molar mass of the species to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._molarMassUnit = v
    molarMassUnit = property(GetMolarMassUnit, SetMolarMassUnit, None, None)

    def GetMolarRadioactivity(self):

        '''
        Returns the numerical value for molar radioactivity of the species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarRadioactivity: float
        '''

        return self._molarRadioactivity

    def SetMolarRadioactivity(self, v):

        '''
        Sets the molar radioactivity of the species equal to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._molarRadioactivity = v
    molarRadioactivity = property( GetMolarRadioactivity, SetMolarRadioactivity,
            None, None)

    def GetMolarRadioactivityFractions(self):

        '''
        Returns a list of numbers that speciefies the % of molar reactivity
        that comes from each type of atom in the species. For example, a
        molarRadioactivityFraction of [0.65, 0.35] for water means that 65%
        of the molar radioactivity comes from the hydrogen atoms and 35% comes
        from the oxygen atom.

        Parameters
        ----------
        empty

        Returns
        -------
        molarRadioactivityFractions: list
        '''

        return self._molarRadioactivityFractions

    def SetMolarRadioactivityFractions(self, fracs):

        '''
        Sets molarRadioactivityFractions equal to fracs. Fracs must be a list
        of floatswith the same length as there are different atoms in the
        species, or the function call will fail. (e.g. self._atoms and fracs
        must be of the same length). Take care to ensure that the elements of
        fracs match with the elements of self._atoms! (65% is in the same
        position in fracs as hydrogen is in self._atoms, following the above
        example).

        Parameters
        ----------
        fracs: list

        Returns
        -------
        empty
        '''

        assert isinstance(fracs, list), 'oops not list.'
        if len(fracs) > 0:
            assert len(fracs) == len(self._atoms), 'oops not right length,'
        if len(fracs) != 0:
            assert isinstance(fracs[-1], float), 'oops not float.'
        self._molarRadioactivityFractions = fracs
    molarRadioactivityFractions = property( GetMolarRadioactivityFractions,
            SetMolarRadioactivityFractions, None, None)

    def GetMolarRadioactivityUnit(self):

        '''
        Returns the unit used to measure molar radioactivity.

        Parameters
        ----------
        empty

        Returns
        -------
        molarRadioactivityUnit: str
        '''

        return self._molarRadioactivityUnit

    def SetMolarRadioactivityUnit(self, v):

        '''
        Sets the unit used to measure molar radioactivity to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._molarRadioactivityUnit = v
    molarRadioactivityUnit = property( GetMolarRadioactivityUnit,
            SetMolarRadioactivityUnit, None, None)

    def GetMolarHeatPwr(self):

        '''
        Returns the amount of heat generated per mole of this species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarHeatPwr: float
        '''

        return self._molarHeatPwr

    def SetMolarHeatPwr(self, v):

        '''
        Sets the amount of heat generated per mole of this species to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._molarHeatPwr = v
    molarHeatPwr = property(GetMolarHeatPwr, SetMolarHeatPwr, None, None)

    def GetMolarHeatPwrUnit(self):

        '''
        Returns the unit used to measure the amount of heat generated per mole
        of this species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarHeatPwrUnit: str
        '''

        return self._molarHeatPwrUnit

    def SetMolarHeatPwrUnit(self, v):

        '''
        Sets the unit used to measure the amount of heat generated per mole of
        this species to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._molarHeatPwrUnit = v
    molarHeatPwrUnit = property( GetMolarHeatPwrUnit, SetMolarHeatPwrUnit, None,
            None)

    def GetMolarGammaPwr(self):

        '''
        Returns the amount of gamma radiation produced per mole of this species
        (measured in units of power).

        Parameters
        ----------
        empty

        Returns
        -------
        molarGammaPwr: float
        '''

        return self._molarGammaPwr

    def SetMolarGammaPwr(self, v):

        '''
        Sets the amount of gamma radiation produced per mole of this species to
        v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._molarGammaPwr = v
    molarGammaPwr = property(GetMolarGammaPwr, SetMolarGammaPwr, None, None)

    def GetMolarGammaPwrUnit(self):

        '''
        Returns the unit used to measure the amount of gamma radiation produced
        per mole of this species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarGammaPwrUnit: str
        '''

        return self._molarGammaPwrUnit

    def SetMolarGammaPwrUnit(self, v):

        '''
        Sets the unit used to measure the amount of gamma radiation produced
        per mole of this species to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._molarGammaPwrUnit = v
    molarGammaPwrUnit = property( GetMolarGammaPwrUnit, SetMolarGammaPwrUnit, None,
            None)

    # Deprecated; see new interface below as GetFormula
    def GetAtoms(self):
        return self._atoms

    def SetAtoms(self, atoms):

        assert isinstance(atoms, list), 'oops not list.'
        if len(atoms) != 0:
            assert isinstance(atoms[-1], str), 'oops not string.'
        self._atoms = atoms
        self.__UpdateMolarMass()
    atoms = property(GetAtoms, SetAtoms, None, None)

    # New interface
    def GetFormula(self):

        '''
        Returns the molecular or empirical formula of the species. It is
        usually a list, for example, of the form ['2*H', 'O'].

        Parameters
        ----------

        Returns
        -------
        formula: list
        '''

        return self._atoms

    def SetFormula(self, atoms):

        '''
        Sets the species' formula equal to atoms. Will automatically update
        the molar mass of the species, and will also fail if atoms is not a
        list of strings.

        Parameters
        ----------
        atoms: list

        Returns
        -------
        empty
        '''

        assert isinstance(atoms, list), 'oops not list.'
        if len(atoms) != 0:
            assert isinstance(atoms[-1], str), 'oops not string.'
        self._atoms = atoms
        self.__UpdateMolarMass()
    formula = property(GetFormula, SetFormula, None, None)

    def GetNAtoms(self):  # number of ficticious atoms in the species (see NB above)

        '''
        Returns the total number of atoms comprising the species. For example,
        water is comprised of three atoms.

        Parameters
        ----------
        empty

        Returns
        -------
        nAtoms: int
        '''

        return self._nAtoms
    nAtoms = property(GetNAtoms, None, None, None)

    # number of nuclide types involved in the species definition
    def GetNNuclideTypes(self):

        '''
        Returns the number of different types of atoms comprising the species.
        For example, water is composed of two different types of atoms,
        hydrogen and oxygen.

        Parameters
        ----------
        empty

        Returns
        -------
        nNuclideTypes: int
        '''

        return self._nNuclideTypes
    nNuclideTypes = property(GetNNuclideTypes, None, None, None)

    def SetFlag(self, f):

        '''
        Sets the flag associated with the species to f.

        Parameters
        ----------
        f: str

        Returns
        -------
        empty
        '''

        self._flag = f

    def GetFlag(self):

        '''
        Returns the flag associated with the species.

        Parameters
        ----------
        empty

        Returns
        -------
        flag: str
        '''

        return self._flag
    flag = property(GetFlag, SetFlag, None, None)

    def GetMolarCC(self):

        '''
        Returns the numerical value for the number (molar) density of the
        species (moles/volume).

        Parameters
        ----------
        empty

        Returns
        -------
        molarCC: float
        '''

        return self._molarCC

    def SetMolarCC(self, v):

        '''
        Sets the numerical value for the molar density of the species to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._molarCC = v
        self._massCC = v * self._molarMass
    molarCC = property(GetMolarCC, SetMolarCC, None, None)

    def GetMolarCCUnit(self):

        '''
        Returns the unit used to measure molar density of the species.

        Parameters
        ----------
        empty

        Returns
        -------
        molarCCUnit: str
        '''

        return self._molarCCUnit

    def SetMolarCCUnit(self, v):

        '''
        Sets the unit used to measure the molar density of the species to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._molarCCUnit = v
    molarCCUnit = property(GetMolarCCUnit, SetMolarCCUnit, None, None)

    def GetMassCC(self):

        '''
        Returns the numerical value of the mass density of the species
        (mass/volume).

        Parameters
        ----------
        empty

        Returns
        -------
        massCC: float
        '''

        return self._massCC

    def SetMassCC(self, v):

        '''
        Sets the numerical value of the mass density equal to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty
        '''

        self._massCC = v
        if self._molarMass == 0.0 and v == 0.0:
            self._molarCC = 0.0
        else:
            self._molarCC = v / self._molarMass
    massCC = property(GetMassCC, SetMassCC, None, None)

    def GetMassCCUnit(self):

        '''
        Returns the unit used to measure the mass density of the species.

        Parameters
        ----------
        empty

        Returns
        -------
        massCCUnit: str
        '''

        return self._massCCUnit

    def SetMassCCUnit(self, v):

        '''
        Sets the units used to measure mass density to v.

        Parameters
        ----------
        v: str

        Returns
        -------
        empty
        '''

        self._massCCUnit = v
    massCCUnit = property(GetMassCCUnit, SetMassCCUnit, None, None)

    def __str__(self):
        s = '\n\t Specie(): name=%s;' + ' formula_name=%s;' + ' phase=%s;' + '\n\t formula=%s;' + '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + ' molar cc=%9.3e[%s];' + ' mass cc=%9.3e[%s];' + '\n\t flag=%s;' + '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t radioactivity  dens.=%9.3e[%s];' + '\n\t molar heat pwr=%9.3e[%s];' + '\n\t heat pwr dens.=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + '\n\t gamma pwr dens.=%9.3e[%s];' + \
            '\n\t atoms=%s;' + '\n\t molar radioactivity fractions=%s'
        return s % (self.name, self.__formula_name, self.phase, self.__ReorderFormula(), self.nAtoms, self.nNuclideTypes, self.molarMass, self.molarMassUnit, self.molarCC, self.molarCCUnit, self.massCC, self.massCCUnit, self.flag, self.molarRadioactivity, self.molarRadioactivityUnit, self.molarRadioactivity *
                    self.molarCC, '[Ci/cc]', self.molarHeatPwr, self.molarHeatPwrUnit, self.molarHeatPwr * self.molarCC, '[W/cc]', self.molarGammaPwr, self.molarGammaPwrUnit, self.molarGammaPwr * self.molarCC, '[W/cc]', [i.split('*')[-1] for i in self.formula], ['%9.3e' % i for i in self.molarRadioactivityFractions])

    def __repr__(self):
        s = '\n\t Specie(): name=%s;' + ' formula_name=%s;' + ' phase=%s;' + '\n\t formula=%s;' + '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + ' molar cc=%9.3e[%s];' + ' mass cc=%9.3e[%s];' + '\n\t flag=%s;' + '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t radioactivity  dens.=%9.3e[%s];' + '\n\t molar heat pwr=%9.3e[%s];' + '\n\t heat pwr dens.=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + '\n\t gamma pwr dens.=%9.3e[%s];' + \
            '\n\t atoms=%s;' + '\n\t molar radioactivity fractions=%s'
        return s % (self.name, self.__formula_name, self.phase, self.__ReorderFormula(), self.nAtoms, self.nNuclideTypes, self.molarMass, self.molarMassUnit, self.molarCC, self.molarCCUnit, self.massCC, self.massCCUnit, self.flag, self.molarRadioactivity, self.molarRadioactivityUnit, self.molarRadioactivity *
                    self.molarCC, '[Ci/cc]', self.molarHeatPwr, self.molarHeatPwrUnit, self.molarHeatPwr * self.molarCC, '[W/cc]', self.molarGammaPwr, self.molarGammaPwrUnit, self.molarGammaPwr * self.molarCC, '[W/cc]', [i.split('*')[-1] for i in self.formula], ['%9.3e' % i for i in self.molarRadioactivityFractions])

#*********************************************************************************
# Private Helper Functions (Internal use: __)
#*********************************************************************************

    def __UpdateMolarMass(self):
        '''
        Updates the molar mass of the species after the molecular formula has
        been changed.

        Parameters
        ----------
        empty

        Returns
        -------
        empty
        '''

        #if len(self._atoms) == 0:
        #    self._nAtoms = 0
        #    return

        for entry in self._atoms:
            assert isinstance(entry, str), 'oops'
            tmp = entry.split('*')
            nuclide = tmp[-1]
            element = nuclide.split('-')[0]
            assert element in ELEMENTS, 'element = %r' % (element)

        self._nAtoms = 0
        self._nNuclideTypes = 0

        nuclides = dict()

        nAtoms = 0
        summ = 0.0
        for entry in self._atoms:
            assert isinstance(entry, str), 'oops'
            # format example:  3.2*O-18, or 3*O or O or O-16
            tmp = entry.split('*')
            multiple = 1.0
            # single nuclide
            if len(tmp) == 1:
                nuclide = tmp[0]
            # multiple nuclide
            elif len(tmp) == 2:
                multiple = float(tmp[0])
                nuclide = tmp[1]
            else:
                assert False

            nuclides[nuclide] = multiple
            nAtoms += multiple

            try:
                tmp = nuclide.split('-')
                if len(tmp) == 1:
                    element = ELEMENTS[tmp[0]]
                    molarMass = element.exactmass  # from isotopic composition
                    if molarMass == 0.0:
                        molarMass = element.mass
                elif len(tmp) == 2:
                    element = ELEMENTS[tmp[0]].isotopes[int(tmp[1].strip('m'))]
                    molarMass = element.mass
                else:
                    assert False
            except KeyError:
                summ += multiple * 0.0
            else:
                summ += multiple * molarMass

        self._molarMass = summ
#   print( summ )
        self._nAtoms = nAtoms
        self._nNuclideTypes = len(nuclides)

        return

    def __ReorderFormula(self):

        '''
        Takes a list of atoms for a molecular or empirical formula and places
        it in order of decreasing magnitude of stoichiometric coefficient. For
        example, [O, 2*H] will be returned as [2*H, O].

        Parameters
        ----------
        empty

        Returns
        -------
        atoms2: list
        '''

        atoms1 = self._atoms[:]  # shallow copy
        atoms2 = list()

        if len(self._atoms) <= 1:
            return atoms1

        if len(self._atoms) > 1:

            # save the multiplier value as a string type of scientific notation
            for entry in self._atoms:

                assert isinstance(entry, str), 'oops'

                # format example:  3.2*O-18, or 3*O or O or O-16
                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                assert multiplier != 0.0, 'multiplier = %r' % (multiplier)

                multiplier = '{0:9.3e}'.format(multiplier)

                atoms1[self._atoms.index(entry)] = multiplier + '*' + tmp[1]

            # order in decreasing order of multiplier magnitude
            multipliers_lst = list()

            for entry in atoms1:

                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                multipliers_lst.append(float(multiplier))

            sortedAtoms_lst = [a for (i, a) in sorted(zip(multipliers_lst, atoms1),
                                                      key=lambda pair: pair[0],
                                                      reverse=True)]

            atoms2 = sortedAtoms_lst

        return atoms2

#============================= end class Specie ==================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Phase *history* container. When you think of a phase value, think of that value at
a specific point in time. This container holds the historic data of a phase;
its species and quantities. This implementation treats access of time stamps within
a tolerance. All searches for time stamped values are subjected to an approximation
of the time stamp to avoid storing values too close to each other in time, and/or to
return the closest value in time searched or no value if none can be found according
to the tolerance.

Background
----------
TODO: ATTENTION:
The species (list of Specie) AND quantities (list of Quantity) data members
have ARBITRARY density values either at an arbitrary point in the history or at
no point in the history. This needs to be removed in the future to avoid confusion.

To obtain history values, associated to the phase, at a particular point in time,
use the GetValue() method to access the history data frame (pandas) via columns and
rows. ALERT: The corresponding values in species and quantities are OVERRIDEN and NOT to
be used through the phase interface.

Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda
Sat Sep  5 01:26:53 EDT 2015

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import sys
from copy import deepcopy
import numpy as npy
import pandas

from cortix.support.specie   import Specie
from cortix.support.quantity import Quantity
#*********************************************************************************

class Phase:
    '''
    Phase `history` container. A `Phase` consists of `Species` and `Quantities`
    varying with time. This container is meant to reproduce the basic idea of a
    material phase.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 time_stamp = None,
                 time_unit  = None,
                 species    = None,
                 quantities = None
                ):
        #TODO
        '''
        Sometimes an empty Phase object is created by user code. This case needs
        adequate logic for None types.
        Note on usage: when passing quantities, do set the value argument explicitly
        to help define the type and avoid SetValue() errors with Pandas. This is
        to be investigated later. Also, the usage of a DataFrame needs to be re-evaluated.
        Maybe better to use a Quantity object and a Specie object with a Pandas Series
        history as a value to avoid the existance of a value in Quantity and a value
        in Phase that are not in sync.
        '''

        if time_stamp is None:
            time_stamp = 0.0 # default type is float
        else:
            assert isinstance(time_stamp, float)
            self.__time_stamp = time_stamp

        if time_unit is None:
            self.__time_unit = 's' # second
        else:
            assert isinstance(time_unit, str)
            self.__time_unit = time_unit

        if species is not None:
            assert isinstance(species, list)
            for specie in species:
                assert isinstance(specie, Specie)

        if quantities is not None:
            assert isinstance(quantities, list)
            for quant in quantities:
                assert isinstance(quant, Quantity)

        # List of species and quantities objects; columns of data frame are named
        # by objects.
        # A new object held by a Phase() object
        self.__species = deepcopy(species)
        # A new object held by a Phase() object
        self.__quantities = deepcopy(quantities)

        names = list()

        if species is not None:
            for specie in self.__species:
                names.append(specie.name)
                specie.massCC = 0.0  # clear these values
                                     # todo: eliminate them from Specie in the future

        if quantities is not None:
            for quant in self.__quantities:
                names.append(quant.name)
                quant.value = 0.0    # clear these values 
                                     # todo: eliminate them from Quantity in the future

        # Table data phase without data type assigned; this is left to the user
        # Time stamps will always be float or int
        self.__phase = pandas.DataFrame( index=[float(time_stamp)], columns=names )

        if species is not None:
            for specie in species:
                self.__phase.loc[time_stamp, specie.name] = specie.molarCC

        if quantities is not None:
            for quant in quantities:
                self.__phase.loc[time_stamp, quant.name] = quant.value
                #self.__phase.fillna( 0.0, inplace=True )  # dtype defaults to float

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def has_time_stamp(self, try_time_stamp):

        '''
        Checks to see if try_time_stamp exists in the phase history.

        Parameters
        ----------
        try_time_stamp:

        Returns
        -------

        '''


        time_stamp = self.__get_time_stamp( try_time_stamp )

        if time_stamp is not None:
            return True
        else:
            return False

    def __get_time_unit(self):
        '''
        Returns the time unit of the `Phase.`

        Parameters
        ----------
        empty

        Returns
        -------
        time_unit: str
        '''

        return self.__time_unit
    time_unit = property(__get_time_unit,None,None,None)

    def GetTimeStamps(self):

        '''
        Returns a list of all the time stamps in the phase history.

        Parameters
        ----------
        empty

        Returns
        -------
        timeStamps: list
        '''

        return list(self.__phase.index)  # return all time stamps
    timeStamps = property(GetTimeStamps, None, None, None)

    def __get_time_stamps(self):

        '''
        Get all time stamps in the index of the data frame.

        Parameters
        ----------
        empty

        Returns
        -------
        time_stamps: list
        '''

        return list(self.__phase.index)  # return all time stamps
    time_stamps = property(__get_time_stamps, None, None, None)

    def GetSpecies(self):

        '''
        Returns every single species in the phase history.

        Parameters
        ----------
        empty

        Returns
        -------
        species: list
        '''

        for species in self.__species:
          tmp = self.GetSpecie(species.name) # handy way to synchronize the whole list
        return self.__species
    species = property(GetSpecies, None, None, None)

    def GetQuantities(self):

        '''
        Returns the list of `Quantities`. The values in each `Quantity` are
        synchronized with the `Phase` data frame.

        Parameters
        ----------
        empty

        Returns
        -------
        quantities: list
        '''

        for quant in self.__quantities:
          tmp = self.GetQuantity(quant.name) # handy way to synchronize the whole list
        return self.__quantities
    quantities = property(GetQuantities, None, None, None)

    def GetActors(self):

        '''
        Returns a list of all the actors in the phase history.

        Parameters
        ----------
        empty

        Returns
        -------
        list(self.__phase.colums): list
        '''

        return list(self.__phase.columns)  # return all names in order

    def GetSpecie(self, name):
        '''
        Returns the species specified by name if it exists, or none if it
        doesn't.

        Parameters
        ----------
        name: str

        Returns
        -------
        specie: str
        '''

        for specie in self.__species:
            if specie.name == name:
                time_stamp = self.__get_time_stamp( None ) # get latest time stamp
                assert name in self.__phase.columns, 'name %r not in %r'% \
                   (name,self.__phase.columns)
                specie.massCC = self.__phase.loc[time_stamp, name]
                return specie  # return specie syncronized with the phase
        return None

    def SetSpecieId(self, name, val):
        '''
        Sets the flag of a specie "name" equal to val.

        Parameters
        ----------
        name: str
        val: int

        Returns
        -------
        empty

        '''

        for specie in self.__species:
            if specie.name == name:
                specie.flag = val
                return

    def GetQuantity(self, name):

        '''
        Returns the quantity evaluated at the last time step of the phase
        history. This also updates the value of the quantity object. If the
        quantity name does not exist the return is None.

        Parameters
        ----------
        name: str

        Returns
        -------
        empty
        '''

        for quant in self.__quantities:
            if quant.name == name:
                time_stamp = self.__get_time_stamp( None ) # get latest time stamp
                assert name in self.__phase.columns, 'name %r not in %r'%\
                   (name,self.__phase.columns)
                quant.value = self.__phase.loc[time_stamp, name]
                return quant  # return quantity syncronized with the phase

        return None

    def get_quantity(self, name, try_time_stamp=None):

        '''
        New version.
        Get the quantity `name` at a point in time closest to
        `try_time_stamp` up to a tolerance. If no time stamp is passed, the
        whole history is returned.

        Parameters
        ----------
        name: str

        try_time_stamp: float, int or None
            Time stamp of desired quantity value. Default: None returns the
            whole quantity history.

        Returns
        -------
        quant.value: float or int or other
        '''

        assert name in self.__phase.columns, 'name %r not in %r'%\
                (name,self.__phase.columns)

        time_stamp = self.__get_time_stamp( try_time_stamp )

        for quant in self.__quantities:
            if quant.name == name:
                quant.value = self.__phase.loc[time_stamp, name] # labels' access mode
                return quant  # return quantity syncronized with the phase

    def get_quantity_history(self, name):

        '''
        Create a Quantity `name` history. This will create a fully qualified
        Quantity object and return to the caller. The function is typically
        needed for data output to a file through `pickle`. Since the value
        attribute of a quantity can be any data structure, a time-series is
        built on the fly and stored in the value attribute. In addition the
        time unit is added to the final return value as a tuple.

        Parameters
        ----------
        name: str

        Returns
        -------
        quant_history: tuple(Quantity,float)
        '''

        assert name in self.__phase.columns, 'name %r not in %r'%\
                (name,self.__phase.columns)

        for quant in self.__quantities:
            if quant.name == name:
                quant_history = deepcopy(quant)
                quant_history.value = self.__phase[name] # whole data frame index series
                return (quant_history,self.__time_unit) # return tuple

    def AddSpecie(self, new_specie):

        '''
        Adds a new specie object to the phase history. See species.py for
        more details on the specie class.

        Parameters
        ----------
        new_specie: obj

        Returns
        -------
        empty
        '''

        assert isinstance(new_specie, Specie)
        assert new_specie.name not in list(self.__phase.columns), \
               'new_specie: %r exists. Current names: %r' % \
               (new_specie, self.__phase.columns)
        speciesFormulae = [specie.formula_name for specie in self.__species]
        assert new_specie.formula_name not in speciesFormulae
        self.__species.append(new_specie)
        newName = new_specie.name
        col = pandas.DataFrame( index=list(self.__phase.index), columns=[newName] )
        tmp = self.__phase
        df = tmp.join(col, how='outer')
        self.__phase = df.fillna(0.0)   # for species have float as default

    def AddQuantity(self, newQuant):

        '''
        Adds a new quantity object to the dataframe. See quantity.py for more
        details on the quantity class.

        Parameters
        ----------
        newQuant: object

        Returns
        -------
        empty
        '''

        assert isinstance(newQuant, Quantity)
        assert newQuant.name not in list(self.__phase.columns), \
               'quantity: %r exists. Current names: %r' % \
               (newQuant, self.__phase.columns)
        quantFormalNames = [quant.formalName for quant in self.__quantities]
        assert newQuant.formalName not in quantFormalNames
        self.__quantities.append(newQuant)
        newName = newQuant.name

        # create a col with object data type; user must fill out column 
        col = pandas.DataFrame( index=list( self.__phase.index), columns=[newName],
                                dtype=object )
        tmp = self.__phase
        df  = tmp.join(col, how='outer')
        #self.__phase = df.fillna(newQuant.value)

    def AddRow(self, try_time_stamp, row_values):


        '''
        Adds a row to the dataframe, with a timestamp of try_time_stamp and
        row values equal to row_values. Take care that the dimensions and order
        of the data matches up!

        Parameters
        ----------
        try_time_stamp: float
        row_values: list

        Returns
        -------
        empty
        '''

        assert try_time_stamp not in self.__phase.index, 'already used time_stamp: %r'%\
                (try_time_stamp)
        assert isinstance(row_values, list)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is None, 'already used time_stamp: %r'%(try_time_stamp)
        time_stamp = try_time_stamp

        assert len(row_values) == self.__phase.columns.size

        # create a row with object data type; users row_values data define data type
        row = pandas.DataFrame( index=[time_stamp],
                                columns=list( self.__phase.columns ), dtype=object )

        for (col,v) in zip(row.columns, row_values):
            row.loc[time_stamp,col] = v

        frames = [self.__phase, row]
        self.__phase = pandas.concat(frames)
        return

    def GetRow(self, try_time_stamp=None):

        '''
        Returns an entire row of the phase dataframe. A row is a series of
        values that are all at the same time stamp.

        Parameters
        ----------
        try_time_stamp: float

        Returns
        -------
        list(self.__phase.loc[time_stamp, :]): list
        '''

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        return list(self.__phase.loc[time_stamp, :])

    def GetColumn(self, actor):

        '''
        Returns an entire column of data. A column is the entire history
        of data associated with a specific actor.

        Parameters
        ----------
        actor: str

        Returns
        -------
        list(self.__phase.loc[:, actor]): list
        '''

        assert isinstance(actor, str)
        assert actor in self.__phase.columns, 'actor %r not in %r'% \
                   (actor,self.__phase.columns)
        return list(self.__phase.loc[:, actor])

    def ScaleRow(self, try_time_stamp, value):

        '''
        Multiplies all of the data in a row (except time stamp) by a scalar
        value.

        Parameters
        ----------
        try_time_stamp: float
        value: float

        Returns
        -------
        empty
        '''

        assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)
        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        assert isinstance(value, int) or isinstance(value, float)
        self.__phase.loc[time_stamp, :] *= value
        return

    def ClearHistory(self, value=0.0):

        '''
        Set species and quantities of history to a given value
        (default to zero value), all time stamps are preserved.

        Parameters
        ----------
        value: float

        Returns
        -------
        empty
        '''

        assert isinstance(value, int) or isinstance(value, float)
        self.__phase.loc[:, :] = value

        return

    def ResetHistory(self, try_time_stamp=None, value=None):

        '''
        Set species and quantities of history to a given value
        (default to zero value) only one time stamp is preserved (default to
        last time stamp).

        Parameters
        ----------
        try_time_stamp: float
        value: float

        Returns
        -------
        empty
        '''

        if value is not None:
           assert isinstance(value, int) or isinstance(value, float) or \
                  isinstance(value, npy.ndarray)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        values = self.GetRow(time_stamp)  # save values

        columns = list(self.__phase.columns)
        assert len(columns) == len(values), 'FATAL: oops internal error.'

        self.__phase = pandas.DataFrame( index=[time_stamp], columns=columns )
        self.__phase.fillna( 0.0, inplace=True )

        if value is None:
            for v in values:
                idx = values.index(v)
                self.__phase.loc[time_stamp, columns[idx]] = v  # restore values
        else:
            self.__phase.loc[time_stamp, :] = value   # set user-given value

        return

    def GetValue(self, actor, try_time_stamp=None):

        '''
        Deprecated: use get_value()
        '''

        assert isinstance(actor, str)
        assert actor in self.__phase.columns, 'actor %r not in %r'% \
                   (actor,self.__phase.columns)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        return self.__phase.loc[time_stamp, actor]

    def get_value(self, actor, try_time_stamp=None):

        '''
        Returns the value associated with a specified actor at a specified
        time stamp.

        Parameters
        ----------
        actor: str
        try_time_stamp: float

        Returns
        -------
        self.__phase.loc[time_stamp, actor]: float
        '''

        assert isinstance(actor, str)
        assert actor in self.__phase.columns, 'actor %r not in %r'% \
                   (actor,self.__phase.columns)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        return self.__phase.loc[time_stamp, actor]

    def SetValue(self, actor, value, try_time_stamp=None):

        '''
        For the record: old def SetValue(self, time_stamp, actor, value):

        Parameters
        ----------
        actor: str
        value: float
        try_time_stamp: float

        Returns
        -------
        empty
        '''

        assert isinstance(actor, str)
        assert actor in self.__phase.columns

        #assert isinstance(value, int) or isinstance(value, float) or \
        #       isinstance(value, npy.ndarray)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        #print('*value =', value)
        #print('*type(value) =', type(value))
        #print('*time_stamp =', time_stamp)
        #print('*actor =', actor)
        #print('*column values =', self.__phase[actor])
        #print('*df =', self.__phase)
        #print('*df.dtypes =', self.__phase.dtypes)
        #print('*df.shape =', self.__phase.shape)
        #print('')
        #if isinstance(value,npy.ndarray):
        #   self.__phase[actor] = self.__phase.astype({actor:type(value)})
        #print('*df.dtypes =', self.__phase.dtypes)
        #print('')

        # Note: user value could have a different type than other column values.
        # If there is a type change, this will not be checked; user has been advised.
        self.__phase.loc[time_stamp, actor] = value

        return

    def set_value(self, actor, value, try_time_stamp=None):
        '''
        New version. Discontinue using SetValue()
        '''
        assert isinstance(actor, str)
        assert actor in self.__phase.columns

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        # Note: user value could have a different type than other column values.
        # If there is a type change, this will not be checked; user has been advised.
        self.__phase.loc[time_stamp, actor] = value

        return

    def WriteHTML(self, fileName):

        '''
        Convert the `Phase` container into an HTML file.

        Parameters
        ---------
        fileName: str

        Returns
        -------
        empty
        '''

        assert isinstance(fileName, str)
        tmp = pandas.DataFrame(self.__phase)
        columnNames = tmp.columns
        speciesNames = [specie.name for specie in self.__species]
        quantityNames = [quantity.name for quantity in self.__quantities]
        for col in columnNames:
            if col in speciesNames:
                idx = speciesNames.index(col)
                specie = self.__species[idx]
                tmp.rename(columns={col: specie.formula_name}, inplace=True)
            elif col in quantityNames:
                idx = quantityNames.index(col)
                quant = self.__quantities[idx]
                tmp.rename( columns={ col: col + '[' + quant.unit + ']'},
                            inplace=True )
            else:
                assert False, 'oops fatal.'

        tmp.to_html(fileName)

        return

    def __str__(self):
        s = '\n\t **Phase()**: \n\t time unit: %s\n\t *quantities*: %s\n\t *species*: %s\n\t *history* #time_stamp=%s\n\t *history end* @%s\n%s'
        return s % (self.__time_unit,self.__quantities, self.__species, len(self.__phase.index),
                    self.__phase.index[-1], self.__phase.loc[self.__phase.index[-1], :])

    def __repr__(self):
        s = '\n\t **Phase()**: \n\t time unit: %s\n\t *quantities*: %s\n\t *species*: %s\n\t *history* #time_stamp=%s\n\t *history end* @%s\n%s'
        return s % (self.__time_unit,self.__quantities, self.__species, len(self.__phase.index),
                    self.__phase.index[-1], self.__phase.loc[self.__phase.index[-1], :])

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __get_time_stamp(self, try_time_stamp=None):

        '''
        Helper method for finding the closest time stamp to `try_time_stamp`
        in the phase history. The pandas index container used for storing
        float data type time stamps will return the nearest time stamp up to a
        tolerance. Whether the time index has one value, this function will
        inspect for the proximity to that value.

        Parameters
        ----------
        try_time_stamp: float, int or None
            Default: None will return the last time stamp.

        Returns
        -------
        self.__phase.index[loc]: float or None
            Will return None if no time stamp within tolerance is found.

        '''

        import numpy as np

        tol = 1.0e-3

        if try_time_stamp is None:
            return self.__phase.index[-1]
        else:
            time_stamps = np.array(self.__phase.index)
            if time_stamps.size >= 2:
               tol = 1.0e-3 * np.diff(time_stamps).mean() # 1e-3 * the mean delta t
            try: # abs(index_value - try_time_stamp) <= tolerance
                loc = self.__phase.index.get_loc( try_time_stamp, method='nearest',
                        tolerance=tol )
            except KeyError: # no value found withing tol
                return None
            else:
                return  self.__phase.index[loc]

#======================= end class Phase =========================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from multiprocessing import Queue

class Port:
    '''
    The Port class provides an interface for creating ports and connecting them to
    other ports for the purpose of data tranfer. Data exchange takes place by
    send and/or receive calls on a given port. The concept of a port is that of a data
    transfer "interaction." This can be one- or two-way with sends and receives.
    A port is connected to only one other port; as two ends of a pipe are connected.
    '''

    def __init__(self, name=None, use_mpi=False):

        self.id = None
        self.name = name
        self.use_mpi = use_mpi

        if self.use_mpi:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.rank = None

        self.q = Queue()    # only passes picke-able objects

        self.connected_port = None # the `other` port connected to `this` port

    def connect(self, port):
        '''
        Connect the port to another port

        `port`: A Port object that represents the port to connect to.
        '''
        assert isinstance(port, Port), "Connecting port must be of Port type"

        self.connected_port = port
        port.connected_port = self
        port.use_mpi = self.use_mpi

    def send(self, data, tag=None):
        '''
        Send data to the connected port

        `data`: Any pickelable form of data
        '''
        if not tag:
            tag = self.id

        if self.connected_port:
            if self.use_mpi:
                self.comm.send(data, dest=self.connected_port.rank, tag=tag)
            else:
                self.q.put(data)

    def recv(self):
        '''
        Returns the data recieved from the connected port.
        NOTE: This function will block until data is received from the connected port
        '''
        if self.connected_port:
            if self.use_mpi:
                return self.comm.recv(source=self.connected_port.rank, tag=self.connected_port.id)
            else:
                return self.connected_port.q.get()

    def __eq__(self, other):
        '''
        Ports are the same if their names are the same
        '''
        if isinstance(other, Port):
            return self.name == other.name

    def __repr__(self):
        '''
        Port name representation
        '''
        return self.name

if __name__ == '__main__':
    # Create some ports
    p1 = Port('test1')
    p2 = Port('test2')

    # Connect the ports
    p1.connect(p2)

    # View connections
    print(p1)
    print(p2)
    print(p1.connected_port)
    print(p2.connected_port)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Fuel segment
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sat Jun 27 14:46:49 EDT 2015
'''
#*********************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random
import pandas

from cortix.support.periodictable import ELEMENTS
from cortix.support.periodictable import SERIES
from cortix.support.specie import Specie
#*********************************************************************************

class FuelSegment():

    # TODO: Species should not be here. Need to replace by Phase instead.
    #      Chopper will be affected

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 geometry = pandas.Series(),
                 species = list()
                 ):

        assert isinstance(geometry, pandas.Series), 'fatal.'
        assert isinstance(species, list), 'fatal.'
        if isinstance(species, list) and len(species) > 0:
            assert isinstance(species[0], Specie)

        self.attribute_names = \
            ['n-segments',
             'fuel-volume',
             'segment-volume',
             'fuel-diameter',
             'fuel-length',
             'mass',
             'mass-dens',
             'mass-cc',
             'nuclides',
             'isotopes',
             'radioactivity',
             'radioactivity-dens',
             'gamma',
             'gamma-dens',
             'heat',
             'heat-dens',
             'molar-heat-pwr',
             'molar-gamma-pwr']

        self.__geometry = geometry
        self.__species  = species

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def get_geometry(self):
        '''
        Returns the geometry of the fuel bundle (cylindrical, hexoganol,
        rectangular, etc).

        Parameters
        ----------
        Empty:

        Returns
        -------
        geometry: str
        '''

        return self.__geometry
    geometry = property(get_geometry, None, None, None)

    def get_species(self):

        '''
        Returns the species object which describes the composition of the fuel
        bundle. The species encapsulates all chemical species present in the
        fuel bundle.
        Parameters
        ----------
        None:

        Returns
        -------
        species: object
        '''

        return self.__species
    species = property(get_species, None, None, None)

    def get_specie(self, name):

        '''
        Returns a specie named [name] from the list of species making up the
        fuel bundle. If no name is specified, this function will return None.


        Parameters
        ----------
        name: str

        Returns
        -------
        specie: obj

        '''

        for specie in self.__species:
            if specie.name == name:
                return specie
        return None
    specie = property(get_specie, None, None, None)

# Get stored fuel segment property either overall or on a nuclide basis
    def get_attribute(self, name, nuclide=None, series=None):
        '''
        Used to get stored fuel segment properties, either overall (as an
        average), or on a nuclide basis. "name" in this case refers to the
        attribute in question. At this point in time, series is not implemented
        and passing it to this function will result in an error. Possible
        attributes that may be retrieved with this function, as well as the
        name to pass to this function to retrieve them are: number of segments
        in the bundle (n-segments, always equal to 1), the id of the segment
        that makes up the bundle (segment-id), the volume of the fuel in the
        bundle  (fuel-volume), the total volume of the segment
        (segment-volume), the diameter (fuel-diameter) and length
        (fuel-length) of the segment, the mass or mass density of the segment
        (mass or mass-cc, respectively), or the total or per-volume
        radioactivity, gamma radiation density or heat density of the fuel
        segment (radioactivity and radioactivityDens, gamma and gamma-dens,
        and heat and heat-dens, respectively).

        Finally, density or total mass of a specific nuclide can be determined
        by passing a specific nuclide to the function, with a name value of
        mass or mass-cc.


        Parameters
        ----------
        name: str
        nuclide: str


        Returns
        -------
        many types

        '''

        attribute_name = name

        assert attribute_name in self.attribute_names, \
               'attribute_name: %r; options: %r; fail.' % \
               (attribute_name, self.attribute_names)

        if nuclide is not None:
            assert isinstance(nuclide,str), 'type(nuclide) = %r' % type(nuclide)
            # no multipliers for now (see below: codes is almost ready for it)
            assert len(nuclide.split('*')) == 1, 'nuclide = %r' % nuclide # sanity check

        if nuclide is not None:
            assert series is None, 'fail.'
        if series is not None:
            assert nuclide is None, 'fail.'

        if series is not None:
            assert False, ' not implemented.'

        if attribute_name == 'isotopes':
            assert nuclide is not None, 'need a nuclide symbol.'

# .................................................................................
# # of segments

        if attribute_name == 'n-segments':
            return 1

# .................................................................................
# segment id 

        if attribute_name == 'segment-id':
            return self.__geometry['segment id']

# .................................................................................
# fuel volume

        if attribute_name == 'fuel-volume':
            return self.__get_fuel_segment_volume()
# .................................................................................
# segment volume

        if attribute_name == 'segment-volume':

            cladding_length = self.__geometry['cladding length [cm]']
            cladding_diam = self.__geometry['OD [cm]']
            volume = cladding_length * math.pi * (cladding_diam / 2.0)**2
            return volume

# .................................................................................
# fuel diameter

        if attribute_name == 'fuel-diameter':

            fuel_diam = self.__geometry['fuel diameter [cm]']
            return fuel_diam

# .................................................................................
# fuel length

        if attribute_name == 'fuel-length':

            fuel_length = self.__geometry['fuel length [cm]']
            return fuel_length

# .................................................................................
# fuel segment overall quantities
        if nuclide is None and series is None:

            # mass or mass concentration
            if attribute_name == 'mass-cc' or attribute_name == 'mass-dens' or attribute_name == 'mass':
                mass_cc = 0.0
                for spc in self.__species:
                    mass_cc += spc.massCC
                if attribute_name == 'mass-cc' or attribute_name == 'mass-dens':
                    return mass_cc
                else:
                    volume = self.__get_fuel_segment_volume()
                    return mass_cc * volume
# radioactivity
            if attribute_name == 'radioactivtyDens' or attribute_name == 'radioactivity':
                radDens = 0.0
                for spc in self.__species:
                    radDens += spc.molarRadioactivity * spc.molarCC
                if attribute_name == 'radioactivity-dens':
                    return radDens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return radDens * volume
# gamma
            if attribute_name == 'gamma-dens' or attribute_name == 'gamma':
                gamma_dens = 0.0
                for spc in self.__species:
                    gamma_dens += spc.molarGammaPwr * spc.molarCC
                if attribute_name == 'gamma-dens':
                    return gamma_dens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return gamma_dens * volume
# heat
            if attribute_name == 'heat-dens' or attribute_name == 'heat':
                heat_dens = 0.0
                for spc in self.__species:
                    heat_dens += spc.molarHeatPwr * spc.molarCC
                if attribute_name == 'heat-dens':
                    return heat_dens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return heat_dens * volume

# .................................................................................
# radioactivity

        if attribute_name == 'radioactivity-dens' or attribute_name == 'radioactivity':
            assert False
            colName = 'Radioactivity Dens. [Ci/cc]'

# .................................................................................
# thermal

        if attribute_name == 'thermalDens' or attribute_name == 'thermal' or  \
           attribute_name == 'heat-dens' or attribute_name == 'heat':
            assert False
            colName = 'Thermal Dens. [W/cc]'

# .................................................................................
# gamma

        if attribute_name == 'gamma-dens' or attribute_name == 'gamma':
            assert False
            colName = 'Gamma Dens. [W/cc]'

# .................................................................................
##########################################################################
# .................................................................................

#  if attribute_name[-4:] == 'Dens' or attribute_name[-2:] == 'CC':
#     attributeDens = True
#  else:
#     attributeDens = False

# .................................................................................
# all nuclide content of the fuel added

#  if nuclide is None and series is None:
#
#     density = 0.0
#
#     density = self.propertyDensities[ colName ].sum()
#
#     if attributeDens is False:
#        volume = self.__get_fuel_segment_volume()
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get chemical element series

#  if series is not None:
#
#     density = 0.0
#
#     for isotope in isotopes:
#       density += self.propertyDensities.loc[isotope,colName]
#
#     if attributeDens is False:
#        volume = self.__get_fuel_segment_volume()
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get specific nuclide (either the isotopes of the nuclide or the specific
# isotope) property: note the most complex case handled is, say: 2*Cs-133.

        if nuclide is not None:

            # a particular nuclide given (atomic number and atomic mass number)
            if len(nuclide.split('-')) == 2:

                nuclideMassNumber = int(nuclide.split('-')[1].strip('m'))
                nuclideSymbol = nuclide.split('-')[0]
                nuclideMolarMass = ELEMENTS[nuclideSymbol].isotopes[nuclideMassNumber].mass

                mass_cc = 0.0

                for spc in self.__species:

                    formula = spc.atoms

                    mole_fraction = 0.0

                    for item in formula:

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            formula_nuclide_symbol = item.split('-')[0].strip()
                            if formula_nuclide_symbol == nuclideSymbol:
                                assert len(item.split('-')) == 2

                            if item.split('*')[0].strip() == nuclide:
                                mole_fraction = 1.0
                            else:
                                mole_fraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            formula_nuclide_symbol = item.split(
                                '*')[1].split('-')[0].strip()
                            if formula_nuclide_symbol == nuclideSymbol:
                                assert len(item.split('*')[1].split('-')) == 2

                            if item.split('*')[1].strip() == nuclide:
                                mole_fraction = float(
                                    item.split('*')[0].strip())
                            else:
                                mole_fraction = 0.0

                        else:
                            assert False

                        mass_cc += spc.molarCC * mole_fraction * nuclideMolarMass

                return mass_cc * self.__get_fuel_segment_volume()

        # chemical element given (only atomic number given)
            elif len(nuclide.split('-')) == 1:

                mass_cc = 0.0

                for spc in self.__species:

                    formula = spc.atoms

                    for item in formula:

                        mole_fraction = 0.0

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            assert len(item.split('-')) == 2
                            formula_nuclide_symbol = item.split('-')[0].strip()
                            formula_nuclide_mass_number = int(
                                item.split('-')[1].strip('m'))
                            formula_nuclide_molar_mass = ELEMENTS[formula_nuclide_symbol].isotopes[formula_nuclide_mass_number].mass

                            if formula_nuclide_symbol == nuclide:
                                mole_fraction = 1.0
                            else:
                                mole_fraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            assert len(item.split('*')[1].split('-')) == 2
                            formula_nuclides_symbol = item.split(
                                '*')[1].split('-')[0].strip()
                            formula_nuclide_mass_number = int(
                                item.split('*')[1].split('-')[1].strip('m'))
                            formula_nuclide_molar_mass = \
                                    ELEMENTS[formula_nuclides_symbol].isotopes[formula_nuclide_mass_number].mass

                            if formula_nuclides_symbol == nuclide:
                                mole_fraction = float(
                                    item.split('*')[0].strip())
                            else:
                                mole_fraction = 0.0

                        else:
                            assert False

                        mass_cc += spc.molarCC * mole_fraction * formula_nuclide_molar_mass

                return mass_cc * self.__get_fuel_segment_volume()

            else:

                assert False

#**********************************************************************************
# Private Helper Functions (internal use: __)
#*********************************************************************************

    def __get_fuel_segment_volume(self):
        '''
        Returns the total volume of fuel in the fuel segment.

        Parameters
        ----------
        empty

        Returns
        -------
        volume: float
        '''

        fuel_length = self.__geometry['fuel length [cm]']
        fuel_diam = self.__geometry['fuel diameter [cm]']

        volume = fuel_length * math.pi * (fuel_diam / 2.0)**2

        return volume

    def __str__(self):

        '''
        Used to print the geometry of the fuel segment and the species that it
        consists of.

        Parameters
        ----------
        empty

        Returns
        -------
        s: str
        '''

        s = 'FuelSegment(): %s\n %s\n'
        return s % (self.__geometry, self.__species)

    def __repr__(self):

        '''
        Used to pront the geometry of the fuel segment and the species that it
        consists of.

        Parameters
        ----------
        empty

        Returns
        -------
        s: str
        '''

        s = 'FuelSegment(): %s\n %s\n'
        return s % (self.__geometry, self.__species)

#============================ end class FuelSegment ==============================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from cortix.src.port import Port

class Module:
    '''
    The representation of a Cortix module.
    This class is to be inherited by every Cortix module.
    It provides facilities for creating modules within the Cortix network.
    '''

    def __init__(self):

        self.use_mpi = False

        self.state = None

        self.ports = []

    def send(self, data, port):
        '''
        Send data through a given port.
        '''

        if isinstance(port, str):
            port = self.get_port(port)
            #matches = [p for p in self.ports if p.name == port]
            #assert len(matches) == 1,\
            #        'matches= %r port= %r, ports= %r'%(matches,port,self.ports)
            #port = matches[0]
        elif isinstance(port, Port):
            assert port in self.ports, "Unknown port!"
        else:
            raise TypeError("port must be of Port or String type")

        port.send(data)

    def recv(self, port):
        '''
        Receive data from a given port
        '''

        if isinstance(port, str):
            port = self.get_port(port)
            #matches = [p for p in self.ports if p.name == port]
            #assert(len(matches) == 1)
            #port = matches[0]
        elif isinstance(port, Port):
            assert port in self.ports, "Unknown port!"
        else:
            raise TypeError("port must be of Port or String type")

        return port.recv()

    def get_port(self, name):
        '''
        Get port by name; if it does not exist, create one.
        '''

        assert isinstance(name, str), 'port name must be of type str'
        port = None

        for p in self.ports:
            if p.name == name:
                port = p
                break

        if port is None:
            reserved_port_names = self._get_reserved_port_names()
            if reserved_port_names is not None:
                assert name in reserved_port_names,\
                        'port name: {}, not allowed by module: {}'.format(name,self)
            port = Port(name,self.use_mpi)
            self.ports.append(port)

        return port

    def connect(self, port_name, connected_port):
        '''
        A simpler interface to create module connectivity. Connect the module port
        with `port_name` to a given `connected_port`.
        '''

        my_port = self.get_port(port_name)
        assert isinstance(connected_port, Port), "Connecting port must be of Port type"
        my_port.connect(connected_port)

    def run(self, state_comm=None, idx_comm=None):
        '''
        Run method.

        Parameters
        ----------
        state_comm: multiprocessing.Queue
            When using the Python `multiprocessing` library `state_comm` must have
            the module's `self.state` in it. That is, `state_comm.put(self.state)`
            must be the last command in the method before `return`.

        idx: index of the state in the communication queue.
        '''
        raise NotImplementedError('Module must implement run()')

    def _get_reserved_port_names(self):
        '''
        This is a helper function to check for reserved names of ports in the module.
        If there is no reservation on names or no checking of reserved names wanted,
        this function is the default function and nothing needs to be done on the
        developer side of the module child class.
        Otherwise this function must be overriden and an iterable container of strings
        must be returned.
        '''

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
This FuelBundle class is a container for usage with other plant-level process modules.
It is meant to represent a fuel bundle of an oxide fuel LWR reactor.
There are three main data structures:

  1) fuel bundle specs
  2) solid phase
  3) gas phase

The container user will have to provide all the data and from then on, this class
will help acess the data.
The printing methods reveal the contained data.

Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sun Dec 27 15:06:55 EST 2015
'''
#*********************************************************************************
import os
import sys
import math
import pandas
from copy import deepcopy
#*********************************************************************************

class FuelBundle():

#*********************************************************************************
# Constructor
#*********************************************************************************

    def __init__(self,
                 specs=pandas.DataFrame()
                 ):

        assert isinstance(specs, pandas.DataFrame), 'oops not pandas table.'

        self.__specs = specs

        self.__solid_phase = None
        self.__gas_phase = None

        return

    # ------
    # Start: Pre-irradiation information

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def get_name(self):

        '''
        Returns the name of the fuel bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        name: str
        '''

        return self.__specs.loc['Name', 1]
    name = property(get_name, None, None, None)

    def get_fuel_enrichment(self):

        '''
        Returns the enrichment of the fuel pins in the bundle, in %.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_enrichment: float
        '''

        return self.__get_fuel_enrichment()
    fuel_enrichment = property(get_fuel_enrichment, None, None, None)

    def get_fresh_u_mass(self):

        '''
        Returns the amount of uranium in the bundle, in grams.

        Parameters
        ----------
        empty

        Returns
        -------
        fresh_u_mass: float
        '''

        return self.__get_fresh_u_mass()
    fresh_u_mass = property(get_fresh_u_mass, None, None, None)

    def get_fresh_u238_mass(self):

        '''
        Returns the amount of uranium-238 in the bucket, in grams.

        Parameters
        ----------
        empty

        Returns
        -------
        fresh_u238_mass: float
        '''

        total_mass = self.__get_fresh_u_mass()
        fuel_enrichment = self.__get_fuel_enrichment()
        return total_mass * (1.0 - fuel_enrichment / 100.0)
    fresh_u238_mass = property(get_fresh_u238_mass, None, None, None)

    def get_fresh_U235_mass(self):

        '''
        Returns the amount of uranium-235 in the bucket, in grams.

        Parameters
        ----------
        empty

        Returns
        -------
        fresh_u235_mass: float
        '''

        total_u_Mass = self.__get_fresh_u_mass()
        fuel_enrichment = self.__get_fuel_enrichment()
        return total_u_mass * fuel_enrichment / 100.0
    fresh_u235_mass = property(get_fresh_U235_mass, None, None, None)

    def get_n_fuel_rods(self):

        '''
        Returns the number of fuel rods in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        n_fuel_rods: int
        '''

        return self.__get_n_fuel_rods()
    n_fuel_rods = property(get_n_fuel_rods, None, None, None)

    # End: Pre-irradiation information
    # ------

    def get_fuel_pin_length(self):

        '''
        Returns the length of each fuel pin in the fuel bundle. A fuel pin is
        a cylindircal section of uranium fuel that is surrounded by cladding.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_pin_length: float
        '''

        return self.__get_fuel_pin_length()

    def set_fuel_pin_length(self, x):

        '''
        Sets the length of all fuel pins in the bundle to x.

        Parameters
        ----------
        empty

        Returns
        -------
        x: float
        '''

        self.__specs.loc['Fuel rods fuel length [in]', 1] = x / 2.54  # in
        return
    fuel_pin_length = property(get_fuel_pin_length, set_fuel_pin_length, None, None)

    def get_fuel_rod_od(self):

        '''
        returns the outer diameter of the fuel rod, in cm.
        A fuel rod consists of a fuel pin surrounded by cladding.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_rod_od: float
        '''

        return self.__get_fuel_rod_od()
    fuel_rod_od = property(get_fuel_rod_od, None, None, None)

    def get_fuel_pin_radius(self):

        '''
        Returns the radius of the fuel pin, in cm.
        Parameters
        ----------

        Returns
        -------

        '''

        return self.__get_fuel_pin_radius()
    fuel_pin_radius = property(get_fuel_pin_radius, None, None, None)

    def get_fuel_pin_volume(self):

        '''
        Returns the volume of fuel in each fuel pin, in cm^3.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_pin_volume: float

        '''

        return self.__get_fuel_pin_volume()
    fuel_pin_volume = property(get_fuel_pin_volume, None, None, None)

    def get_fuel_volume(self):
        '''
        Returns the total volume of fuel in the bundle, in cm^3.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_volume: float
        '''

        return self.__get_fuel_volume()
    fuel_volume = property(get_fuel_volume, None, None, None)

    # mass of the solid phase (gas phase in plenum not added)
    def get_fuel_mass(self):

        '''
        Returns the total numerical value for  mass of fuel in the solid phase
        in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_mass: float
        '''

        return self.__solid_phase.GetValue('mass')
    fuel_mass = property(get_fuel_mass, None, None, None)

    def get_fuel_mass_unit(self):

        '''
        Returns the unit used to measure the mass of fuel in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_mass_unit: str
        '''

        return self.__solid_phase.GetQuantity('mass').unit
    fuel_mass_unit = property(get_fuel_mass_unit, None, None, None)

    def get_gas_mass(self):

        '''
        Returns the total numerical value for mass of the fuel in the gas
        phase.

        Parameters
        ----------

        Returns
        -------

        '''

        return self.__gas_phase.GetValue('mass')
    gas_mass = property(get_gas_mass, None, None, None)

    def get_radioactivity(self):

        '''
        Returns the total radioactivity of the fuel bundle, in curies.

        Parameters
        ----------
        empty

        Returns
        -------
        raduioactivity: float
        '''

        return self.__solid_phase.GetValue('radioactivity') + \
            self.__gas_phase.GetValue('radioactivity')
    radioactivity = property(get_radioactivity, None, None, None)

    def get_gamma_pwr(self):

        '''
        Returns the total amount of gamma radiation given by the fuel bundle,
        in watts.

        Parameters
        ----------
        empty

        Returns
        -------
        gamma_pwr: float
        '''

        return self.__solid_phase.GetValue('gamma') + \
            self.__gas_phase.GetValue('gamma')
    gamma_pwr = property(get_gamma_pwr, None, None, None)

    def get_heat_pwr(self):

        '''
        Returns the total amount of heat produced by the fuel bundle, in watts.

        Parameters
        ----------
        empty

        Returns
        -------
        heat_pwr: float
        '''

        return self.__solid_phase.GetValue('heat') + \
            self.__gas_phase.GetValue('heat')
    heat_pwr = property(get_heat_pwr, None, None, None)

    def get_fuel_radioactivity(self):
    # radioactivity of the solid phase

        '''
        Returns the total radioactivity of the fuel in the solid phase in the
        fuel bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_radioactivity: float
        '''

        return self.__solid_phase.GetValue('radioactivity')
    fuel_radioactivity = property(get_fuel_radioactivity, None, None, None)

    def get_gas_radioactivity(self):
        # radioactivity of the gas phase

        '''
        Returns the total radioactivity of the fuel in the gas phase in the
        fuel bundle, in curies.

        Parameters
        ----------
        empty

        Returns
        -------
        gas_radioactivity: float
        '''

        return self.__gas_phase.GetValue('radioactivity')
    gas_radioactivity = property(get_gas_radioactivity, None, None, None)

    def get_solid_phase(self):

        '''
        Returns the solid phase history associated with this fuel bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        solidPhase: dataFrame
        '''

        return self.__solid_phase

    def set_solid_phase(self, phase):

        '''
        Sets the solid phase history of the fuel equal to phase.

        Parameters
        ----------
        phase: dataFrame

        Returns
        -------
        empty
        '''

        self.__solid_phase = deepcopy(phase)
    solid_phase = property(get_solid_phase, set_solid_phase, None, None)

    def get_gas_phase(self):

        '''
        Returns the gas phase history of the fuel.

        Parameters
        ----------
        empty

        Returns
        -------
        gas_phase: dataFrame
        '''

        return self.__gas_phase

    def set_gas_phase(self, phase):

        '''
        Sets the gas phase history of the fuel equal to phase.

        Parameters
        ----------
        phase: dataFrame

        Returns
        -------
        empty
        '''

        self.__gas_phase = deepcopy(phase)
    gas_phase = property(get_gas_phase, set_gas_phase, None, None)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __get_fuel_enrichment(self):

        '''
        Returns the enrichment of the fuel.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc['Enrichment [U-235 wt%]', 1]: float
        '''

        return float(self.__specs.loc['Enrichment [U-235 wt%]', 1])

    def __get_fresh_u_mass(self):
        '''
        Returns the mass of fuel in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc['U mass per assy [kg]', 1]) * 1000
        '''

        return float(
            self.__specs.loc['U mass per assy [kg]', 1]) * 1000.0  # [g]

    def __get_n_fuel_rods(self):

        '''
        Returns the number of fuel rods in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc['Fuel rods number', 1]): int
        '''

        return int(self.__specs.loc['Fuel rods number', 1])

    def __get_fuel_pin_length(self):

        '''
        Returns the length of the fuel pins in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc["fuel rods fuel length [in]', 1]) * 2.54
        '''

        return float(
            self.__specs.loc['Fuel rods fuel length [in]', 1]) * 2.54  # cm

    def __get_fuel_rod_od(self):

        '''
        Returns the outer diameter of the fuel rods in the bundle.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc['Fuel rods O.D. [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Fuel rods O.D. [in]', 1]) * 2.54  # cm

    def __get_fuel_rod_wall_thickness(self):

        '''
        Returns the thickness of the cladding wall on the outside of the fuel
        rod.

        Parameters
        ----------
        empty

        Returns
        -------
        self.__specs.loc['Fuel rods wall thickness [in]', 1]) * 2.54: float
        '''

        return float(
            self.__specs.loc['Fuel rods wall thickness [in]', 1]) * 2.54  # cm

    def __get_fuel_pin_radius(self):

        '''
        Returns the radius of the fuel pins.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_pin_radius: float
        '''

        fuel_rod_od = self.__get_fuel_rod_od()
        fuel_rod_wall_thickness = self.__get_fuel_rod_wall_thickness()
        fuel_pin_radius = (fuel_rod_od - 2.0 * fuel_rod_wall_thickness) / 2.0
        return fuel_pin_radius

    def __get_fuel_pin_volume(self):

        '''
        Returns the volume of each fuel pin.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_pin_length * math.pi * fuel_pin_radius ** 2: float
        '''

        fuel_pin_length = self.__get_fuel_pin_length()
        fuel_pin_radius = self.__get_fuel_pin_radius()
        return fuel_pin_length * math.pi * fuel_pin_radius ** 2

    def __get_fuel_volume(self):
        '''
        Returns the volume of fuel in the bundle, in cm^3.

        Parameters
        ----------
        empty

        Returns
        -------
        fuel_pin_volume * n_fuel_rods: float
        '''

        fuel_pin_volume = self.__get_fuel_pin_volume()
        n_fuel_rods = self.__get_n_fuel_rods()
        return fuel_pin_volume * n_fuel_rods

    def __str__(self):
        s = 'FuelBundle():\n %s\n %s\n %s\n'
        return s % (self.__specs, self.__solid_phase, self.__gas_phase)

    def __repr__(self):
        s = 'FuelBundle():\n %s\n %s\n %s\n'
        return s % (self.__specs, self.__solid_phase, self.__gas_phase)

#======================= end class FuelBundle ====================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda

Stream container

VFdALib support classes

Sat Aug 15 17:24:02 EDT 2015
'''
#*********************************************************************************
import os
import sys
import pandas

from cortix.support.specie import Specie
from cortix.support.quantity import Quantity
#*********************************************************************************

class Stream:

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 timeStamp,
                 species=None,
                 quantities=None,
                 values=float(0.0)
                 ):

        # constructor
        # THIS WILL CREATE AN ORDERED "STREAM"; values must be in order!

        assert isinstance(timeStamp, float)

        self.timeStamp = timeStamp

        if species is not None:
            assert isinstance(species, list)
            if len(species) > 0:
                assert isinstance(species[-1], Specie)

        if quantities is not None:
            assert isinstance(quantities, list)
            if len(quantities) > 0:
                assert isinstance(quantities[-1], Quantity)

#  assert type(value) == type(float())

# List of quantities and species objects
        self.species = species       # list
        self.quantities = quantities    # list

        names = list()  # note order of names; values must be in the same order; caution

        for specie in self.species:
            names.append(specie.name)

        for quant in self.quantities:
            names.append(quant.name)

# ORDERED data; caution!!
# Table data stream
        self.stream = pandas.DataFrame(index=[timeStamp], columns=names)
        if isinstance(values, float):
            self.stream.fillna(values, inplace=True)
        else:
            self.stream.fillna(0.0, inplace=True)

        if isinstance(values, list) and len(values) == len(names):
            for (name, val) in zip(names, values):
                self.stream.loc[timeStamp, name] = val

        return

#*********************************************************************************
# Public Member Functions
#*********************************************************************************

    def GetTimeStamp(self):

        '''
        Returns the time stamp of the stream.

        Parameters
        ----------
        empty

        Returns
        -------
        self.timeStamp: float
        '''

        return self.timeStamp

    def GetActors(self):

        '''
        Returns the actors present in the stream of data.

        Parameters
        ----------
        empty

        Returns
        -------
        list(self.stream.columns): list
        '''

        return list(self.stream.columns)

    def GetSpecie(self, name):

        '''
        Returns a specie named "name" from the stream.

        Parameters
        ----------
        name: str

        Returns
        -------
        specie: obj
        '''

        for specie in self.species:
            if specie.name == name:
                return specie
        return None

    def GetSpecies(self):

        '''
        Returns a list of all species in the stream.

        Parameters
        ----------
        empty

        Returns
        -------
        self.species: list
        '''

        return self.species

    def GetQuantities(self):

        '''
        Returns all the quantities given by the stream.

        Parameters
        ----------
        empty

        Returns
        -------
        self.quantities: list
        '''

        return self.quantities

    def SetSpecieId(self, name, val):

        '''
        Sets the numerical id of the specie of name "name" to val.

        Parameters
        ----------
        name: str
        val: int

        Returns
        -------
        empty
        '''

        for specie in self.species:
            if specie.name == name:
                specie.flag = val
                return

    def GetQuantity(self, name):

        '''
        Returns the specified quantity called "name" from the stream, or none
        if the specified name does not exist.

        Parameters
        ----------
        name: str

        Returns
        -------
        quant: float
        '''

        for quant in self.quantities:
            if quant.name == name:
                return quant
        return None

    def GetRow(self, timeStamp=None):

        '''
        Returns an entire row of data from the stream. A row of data is all
        the data in a dataframe at a specified time stamp, given by timeStamp.
        If timeStamp is not specified, this function will return the entire
        stream dataframe.

        Parameters
        ----------
        timeStamp: float

        Returns
        -------
        self.stream.loc[self.timestamp, :]) or self.stream.loc[timeStamp, :]):
        list
        '''

        if timeStamp is None:
            return list(self.stream.loc[self.timeStamp, :])
        else:
            assert timeStamp in self.stream.index, 'timeStamp = %r' % (
                timeStamp)
            return list(self.stream.loc[timeStamp, :])

    def GetValue(self, actor, timeStamp=None):

        '''
        Returns the value associated with a specified "actor" at a specified
        "timeStamp". If no timeStamp is specified, then the function will
        return all values associated with the specified actor at all time
        stamps.

        Parameters
        ----------
        actor: str
        timeStamp: float

        Returns
        -------
        self.stream.loc[self.timeStamp, actor] or self.stream.loc[timeStamp,
        actor]: list or float, respectively.
        '''

        assert actor in self.stream.columns
        if timeStamp is None:
            return self.stream.loc[self.timeStamp, actor]
        else:
            assert timeStamp in self.stream.index
            return self.stream.loc[timeStamp, actor]

    def SetValue(self, actor, value=None, timeStamp=None):

        '''
        Sets the value associated with a specified actor at a specified
        timeStamp to "value". If no value is specified, the value will default
        to 0.0. If no timeStamp is specified, it will set all values associated
        with actor to the specified value (or 0.0 if value = None).

        Parameters
        ----------
        actor: str
        value: float
        timeStamp: float

        Returns
        -------
        empty
        '''

        assert actor in self.stream.columns
        if timeStamp is None:
            if value is None:
                self.stream.loc[self.timeStamp, actor] = float(0.0)
            else:
                self.stream.loc[self.timeStamp, actor] = float(value)
        else:
            assert timeStamp in self.stream.index, 'timeStamp = %r' % (
                timeStamp)
            if value is None:
                self.stream.loc[timeStamp, actor] = float(0.0)
            else:
                self.stream.loc[timeStamp, actor] = float(value)

# def __str__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
#
# def __repr__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

#============================= end class Stream ==================================



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    req = fh.readlines()

setuptools.setup(
    name="cortix",
    version="1.0.3",
    author="Cortix Computing",
    author_email="tazzaoui@cs.uml.edu",
    description="Cortix is a Python library for system-level\
                module coupling, execution, and analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=req,
    url="https://cortix.org",
    packages=setuptools.find_namespace_packages(),
    keywords = ['simulation', 'math'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Topic :: Scientific/Engineering :: Mathematics',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Topic :: Education',
        'Topic :: Utilities'
    ],
)



#!/usr/bin/env python

from cortix.src.module import Module
from cortix.examples.dummy_module import DummyModule

from cortix.src.cortix_main import Cortix

def test_cortix_add_module():
    # Init the Cortix object
    c = Cortix()

    num_modules = 100
    module_list = list()

    # Add 100 modules to the Cortix object
    for i in range(num_modules):
        # Initialize the module
        m = DummyModule()
        c.add_module(m)

        # Get ports
        p1 = m.get_port('test1-{}'.format(i))
        p2 = m.get_port('test2-{}'.format(i))

    # Make sure we have the correct modules
    assert len(c.modules) == num_modules
    for mod in c.modules:
        assert isinstance(mod, Module)
        assert len(mod.ports) == 2

if __name__ == "__main__":
    test_cortix_add_module()



from cortix.src.module import Module
from cortix.src.port import Port
from cortix.examples.dataplot import DataPlot
from cortix.src.cortix_main import Cortix

class PlotData(Module):
    def __init__(self):
        super().__init__()

    def run(self):
        i = 0
        while i < 10:
            data = (i, i**2)
            self.send(data, "plot-out")
            print("Sent {}!".format(data))
            i += 1
        self.send("DONE", "plot-out")
        print("Finished sending!")

if __name__ == "__main__":

    # Cortix built-in DataPlot module
    d = DataPlot()
    d.set_title("Test Plot")
    d.set_xlabel("Time")
    d.set_ylabel("Position")

    # Custom class to send dummy data
    p = PlotData()

    p1 = Port("plot-in")
    p2 = Port("plot-out")
    p1.connect(p2)

    d.add_port(p1)
    p.add_port(p2)

    c = Cortix()
    c.add_module(d)
    c.add_module(p)
    c.use_mpi = False
    c.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

This Quantity class is to be used with other classes in plant-level process modules.

For unit testing do at the linux command prompt:
    python quantity.py

Sat Sep  5 12:51:34 EDT 2015
'''
#*********************************************************************************
import os
import sys
import numpy as npy
#*********************************************************************************

class Quantity:
    '''
    todo: this probably should not have a "value" for the same reason as Specie.
          this needs some thinking.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 name       = 'null-quantity',
                 formalName = 'null-quantity',
                 value      = float(0.0),  # this can be any type
                 unit       = 'null-unit'
                ):

        # Sanity tests here
        assert isinstance(name, str), 'oops not string.'
        self._name = name

        assert isinstance(formalName, str), 'oops not string.'
        self._formalName = formalName

        self.__value = value

        assert isinstance(name, str), 'oops not string.'
        self.__unit = unit

        self.__name = name
        self._formalName = formalName
        self.__value = value
        self.__unit = unit

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def SetName(self, n):

        '''
        Sets the name of the quantity in question to n.

        Parameters
        ----------
        n: str

        Returns
        -------
        empty
        '''
        self.__name = n

    def get_name(self):

        '''
        Returns the name of the quantity.

        Parameters
        ----------
        empty

        Returns
        -------
        name: str
        '''

        return self.__name
    name = property(get_name, SetName, None, None)

    def SetValue(self, v):

        '''
        Sets the numerical value of the quantity to v.

        Parameters
        ----------
        v: float

        Returns
        -------
        empty

        '''
        self.__value = v

    def GetValue(self):

        '''
        Gets the numerical value of the quantity.

        Parameters
        ----------
        empty

        Returns
        -------
        value: float
        '''

        return self.__value
    value = property(GetValue, SetValue, None, None)

    def SetFormalName(self, fn):
        '''
        Sets the formal name of the property to fn.

        Parameters
        ----------
        fn: str

        Returns
        -------
        empty
        '''

        self._formalName = fn

    def GetFormalName(self):

        '''
        Returns the formal name of the quantity.

        Parameters
        ----------
        empty

        Returns
        -------
        formalName: str
        '''

        return self._formalName
    formalName = property(GetFormalName, SetFormalName, None, None)
    formal_name = property(GetFormalName, SetFormalName, None, None)

    def SetUnit(self, f):

        '''
        Sets the units of the quantity to f (for example, density would be in
        units of g/cc.

        Parameters
        ----------
        f: str

        Returns
        -------
        empty
        '''

        self.__unit = f

    def GetUnit(self):

        '''
        Returns the units of the quantity.

        Parameters
        ----------
        empty

        Returns
        -------
        unit: str
        '''

        return self.__unit
    unit = property(GetUnit, SetUnit, None, None)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __str__(self):

        '''
        Used to print the data stored by the quantity class. Will print out
        name, formal name, the value of the quantity and its unit.

        Parameters
        ----------
        empty

        Returns
        -------
        s: str
        '''

        s = '\n\t Quantity(): \n\t name=%s; formal name=%s; value=%s[%s]'
        return s % (self.name, self.formalName, self.value, self.unit)

    def __repr__(self):

        '''
        Used to print the data stored by the quantity class. Will print out
        name, formal name, the value of the quantity and its unit.

        Parameters
        ----------
        empty

        Returns
        -------
        s: str
        '''

        s = '\n\t Quantity(): \n\t name=%s; formal name=%s; value=%s[%s]'
        return s % (self.name, self.formalName, self.value, self.unit)

#======================= end class Quantity ======================================
